open Ast


module NameMap = Map.Make(struct
  type t = string
  let compare x y = Pervasives.compare x y
end)

exception ReturnException of string NameMap.t * string NameMap.t NameMap.t



									(* Main entry point: run a program *)

let run (vars, funcs) =

									(* Put function declarations in a symbol table *)
   let func_decls = List.fold_left
      (fun funcs fdecl -> 
	if NameMap.mem fdecl.fname funcs then 
	   raise (Failure ("function " ^ fdecl.fname ^ " is defined more than once!"));
	NameMap.add fdecl.fname fdecl funcs ) NameMap.empty funcs
   in


									(* function for initializing variables *)
   let init_variables = List.fold_left
      (fun globals vdecl -> 

	let params = List.fold_left  
		(fun param_map param_decl -> NameMap.add param_decl.p_key param_decl.p_value param_map)
		NameMap.empty vdecl.v_params in

	if NameMap.mem vdecl.v_name globals then 
	   	raise (Failure ("variable " ^ vdecl.v_name ^ " is defined more than once!"));

        NameMap.add vdecl.v_name params globals; 

      ) 

   in


   									(* Invoke a function and return an updated global symbol table *)
   let rec call fdecl actuals globals =

	

									(* Evaluate an lexpr and return (value, updated environment) *)
	let rec l_eval env = function

             Literal(i) -> i, env 

	   | StrConst(i) -> (String.sub i 1 ((String.length i)-2)),env

	   | Param(var, param) ->

		let (locals, globals) = env in

		if NameMap.mem var locals then
		    if NameMap.mem param (NameMap.find var locals) then
			NameMap.find param (NameMap.find var locals),(locals, globals)
		    else raise (Failure ("undeclared identifier " ^ var))

		else if NameMap.mem var globals then
		    if NameMap.mem param (NameMap.find var globals) then
			NameMap.find param (NameMap.find var globals),(locals, globals)
	            else raise (Failure ("undeclared identifier " ^ var))
		else raise (Failure ("undeclared identifier " ^ var))

	   | Uminus(e1) ->
		let v1, env = l_eval env e1 in
		string_of_float (-.(float_of_string v1)),env

	   | Binop(e1, op, e2) ->
		let v1, env = l_eval env e1 in
		let v2, env = l_eval env e2 in
		let boolean_to_str i = if i then "yes" else "no" in
		let int_to_str i = if i=0 then "yes" else "no" in
		(match op with
		  Add  -> string_of_float ((float_of_string v1) +. (float_of_string v2)) 
		| Sub  -> string_of_float ((float_of_string v1) -. (float_of_string v2))
		| Mult -> string_of_float ((float_of_string v1) *. (float_of_string v2))
		| Div  -> string_of_float ((float_of_string v1) /. (float_of_string v2))
		| Equal-> boolean_to_str  ((float_of_string v1) =  (float_of_string v2))	
		| Neq  -> boolean_to_str  ((float_of_string v1) <> (float_of_string v2)) 
		| Less -> boolean_to_str  ((float_of_string v1) <  (float_of_string v2)) 
		| Leq  -> boolean_to_str  ((float_of_string v1) <= (float_of_string v2))
		| Greater -> boolean_to_str ((float_of_string v1) >  (float_of_string v2))
		| Geq -> boolean_to_str   ((float_of_string v1) >= (float_of_string v2))
		| Concat -> String.concat "" (v1::(v2::[])) 
		| Cmp ->  int_to_str (String.compare v1 v2)  
		| And ->  if (0 = (String.compare v1 "yes") && 
			      0 = (String.compare v2 "yes")) then "yes" else "no"
		| Or ->   


		if (0 = (String.compare v1 "yes") || 
			      0 = (String.compare v2 "yes")) then "yes" else "no"  ), env

	    | AssignL(var, param, e) ->

		let v, (locals, globals) = l_eval env e in

		if NameMap.mem var locals then
			v, (NameMap.add var (NameMap.add param v (NameMap.find var locals) ) locals, globals)
		else if NameMap.mem var globals then
			v, (locals, NameMap.add var (NameMap.add param v (NameMap.find var globals) ) globals) 
		else raise (Failure ("undeclared identifier " ^ var))

	in


     									(* Evaluate an expression and return (value, updated environment) *)
	let rec eval env = function

	     Id(var) ->

		let locals, globals = env in
		if NameMap.mem var locals then
			(NameMap.find var locals), env
		else if NameMap.mem var globals then
			(NameMap.find var globals), env 
		else raise (Failure ("undeclared identifier " ^ var))

	   | Assign(var, param, e) ->

		let v, (locals, globals) = l_eval env e in

		if NameMap.mem var locals then
			NameMap.empty, (NameMap.add var (NameMap.add param v (NameMap.find var locals) ) locals, globals)
		else if NameMap.mem var globals then
			NameMap.empty, (locals, NameMap.add var (NameMap.add param v (NameMap.find var globals) ) globals) 
		else raise (Failure ("undeclared identifier " ^ var))

	   | AssignVar(var, e) ->

		let v, (locals, globals) = eval env e in
		
		if NameMap.mem var locals then
			v,((NameMap.add var v locals),globals)
		else if NameMap.mem var globals then
			v,(locals,(NameMap.add var v globals))
		else raise (Failure ("undeclared identifier " ^ var))


	   | Call("printAll", []) ->

		let print_params key value = print_string "  "; print_string key; print_endline "->";
					     NameMap.iter (fun k v -> print_string "     "; 
							    print_string k; 
							    print_string " = ";
							    print_endline v;) value in
		print_endline "------------------------";
		print_endline ("-Global-");
		NameMap.iter print_params (snd env);
		print_endline "";
		print_endline ("-Local-");
		NameMap.iter print_params (fst env);
		print_endline "------------------------";
		NameMap.empty, env


	   | Call("print", [var_list]) ->


		let var, env = eval env var_list in 
			
		print_endline "";
		print_endline ("-Listing-");
		NameMap.iter (fun k v -> print_string "     "; 
					 print_string k; 
					 print_string " = ";
					 print_endline v;) var;

		NameMap.empty, env

	   | Call(f, actuals) ->

		let fdecl = 
		    try NameMap.find f func_decls
		    with Not_found -> raise (Failure ("undefined function " ^ f)) in

		let actuals, env = 
		    List.fold_left (fun (actuals, env) actual ->
				    let v, env = eval env actual in v :: actuals, env)
			 ([], env) actuals in

		let (locals, globals) = env in

		    try 
			let globals = call fdecl actuals globals in
			   NameMap.empty, (locals, globals)
		    with ReturnException(v, globals) ->
		           v, (locals, globals)  
       in
									

							(* Execute a statement and return an updated environment *)
       let rec exec env = function

	  Block(stmts) -> List.fold_left exec env stmts

	| Expr(e) -> let _, env = eval env e in env

	| If(e, s1, s2) -> let v, env = l_eval env e in
			exec env (if String.compare "yes" v == 0 then s1 else s2)

	| While(e, s) -> let rec loop env =
			   let v, env = l_eval env e in
				if String.compare "yes" v == 0 then loop (exec env s) 
				else env
			in loop env

	| Return(e) -> let v, (locals, globals) = eval env e in
			raise (ReturnException(v, globals))

       in


								(* call: enter the function: bind actual values to formal args *)
       let locals =
          try List.fold_left2
             (fun locals formal actual -> NameMap.add formal actual locals)
             NameMap.empty fdecl.formals actuals
          with Invalid_argument(_) ->
              raise (Failure ("wrong number of arguments to " ^ fdecl.fname))
       in
  
								(* Initilize local variables *)
       let locals = init_variables locals fdecl.locals in 


								(* Execute each statement; return updated global symbol table *)
       snd (List.fold_left exec (locals, globals) fdecl.body)
									

   in 
								(* add global variables to symbol table. *)

   let globals = init_variables NameMap.empty vars in 

								(* Run the "program" function *)
   try
      call (NameMap.find "program" func_decls) [] globals;
   with Not_found ->
      raise (Failure ("did not find the program() function"))



